"""
Author: Taylor B. tayjaybabee@gmail.com
Date: 2024-11-23 13:09:11
LastEditors: Taylor B. tayjaybabee@gmail.com
LastEditTime: 2024-11-27 20:23:35
FilePath: inspy_logger/engine/adapters/thread.py
Description: 这是默认设置,可以在设置》工具》File Description中进行配置
"""
import logging
import threading


class ThreadAdapter(logging.LoggerAdapter):
    """
    A class that adapts a logger to work with threads.
    """
    def process(self, msg, kwargs):
        """
        Processes the message and keyword arguments.

        Parameters:
            msg (str): The message to log.
            kwargs (Dict): The keyword arguments to log.

        Returns:
            Tuple: A tuple containing the message and keyword arguments.
        """
        # Add context to the log message.
        thread_name = threading.current_thread().name

        if thread_name != 'MainThread':
            msg = f'[Thread - {thread_name}] {msg}'

        return msg, kwargs



import inspect
import os
import logging
import sys

from time import time

from rich.logging import RichHandler

from inspy_logger.config import DEFAULT_LOG_FILE_PATH
from inspy_logger.constants import LEVELS, INTERACTIVE_SESSION, INTERNAL, HANDLER_TYPES
from inspy_logger.engine.handlers import BufferingHandler
from inspy_logger.models.announcement import Announcement
from inspy_logger.common import InspyLogger, DEFAULT_LOGGING_LEVEL
from inspy_logger.helpers import (
    translate_to_logging_level, CustomFormatter, get_level_name,
    translate_to_logging_level_str
    )
from inspy_logger.helpers.decorators import add_aliases, method_alias, count_invocations, validate_type, \
    property_logging
from typing import List, Union, Optional
from pathlib import Path
from warnings import warn


@add_aliases
class Logger(InspyLogger):
    """
    A Singleton class responsible for managing the logging mechanisms of the application.
    """

    LEVELS = LEVELS
    INTERACTIVE_SESSION = INTERACTIVE_SESSION

    TRACE_LEVEL_NUM = 5

    instances = {}  # A dictionary to hold instances of the Logger class.

    def __new__(cls, name, *args, **kwargs):
        """
        Creates or returns an existing instance of the Logger class for the provided name.

        Parameters:
            name (str):
                The name of the logger instance.

        Returns:
            Logger:
                An instance of the Logger class.
        """

        if name not in cls.instances:
            instance = super().__new__(cls)
            cls.instances[name] = instance
            return instance
        return cls.instances[name]

    def __init__(
            self,
            name,
            auto_set_up=True,
            console_level=DEFAULT_LOGGING_LEVEL,
            file_level=logging.DEBUG,
            file_name="app.log",
            file_path=DEFAULT_LOG_FILE_PATH.parent,
            no_file_logging=False,
            parent=None,
            init_announcement: Optional[Announcement] = None,
            init_announcement_template: str = None,
            announce_on_init: bool = True,
            announcement_level: Union[int, str] = 'debug'
            ):
        """
        Initializes a logger instance.

        Parameters:
            name (str):
                The name of the logger instance.

            auto_set_up (bool, optional):
                Whether to automatically set up the handlers for the logger. Defaults to True.

            console_level (str, optional):
                The logging level for the console. Defaults to DEFAULT_LOGGING_LEVEL.

            file_level (str, optional):
                The logging level for the file. Defaults to logging.DEBUG.

            file_name (str, optional):
                The name of the log file. Defaults to "app.log".

            no_file_logging (bool, optional):
                Whether to disable file logging. Defaults to False.

            parent (Logger, optional):
                The parent logger instance. Defaults to None.

            init_announcement (Announcement, optional):
                An announcement object to use for initialization. Defaults to None.

            init_announcement_template (str, optional):
                A template to use for the initialization announcement. Defaults to None.

                Note:
                    If both `init_announcement` and `init_announcement_template` are provided, `init_announcement` will
                    take precedence.

                    Built-In Placeholders:
                        - {name}: The name of the logger.
                        - {time_started}: The time the logger was started.
                        - {parent}: The name of the parent logger.


        """
        # Check if the logger has already been initialized.
        if hasattr(self, 'logger'):
            return

        self.__time_started = time()

        self.__announcement_made = False

        self.__call_counts = {}
        self.__console_level = translate_to_logging_level(console_level)
        self.__file_level = translate_to_logging_level(file_level)

        self.__children = []

        self.__name = name
        self.__no_file_logging = None
        self.__file_path = None
        self.__warnings_issued = set()

        self.logger = logging.getLogger(name)

        self.logger.setLevel(translate_to_logging_level(console_level))

        self.logger.propagate = False

        self.parent = parent

        self.logger.start = self.start

        if 'inSPy-Logger' in self.logger.name:
            self.buffering_handler = BufferingHandler()
            self.logger.addHandler(self.buffering_handler)
            self.internal('Initializing logger with buffering handler.')
        else:
            self.internal('Initializing  logger without buffering handler.')

        self.no_file_logging = no_file_logging

        self._file_path = Path(file_path).expanduser().absolute().joinpath(file_name)

        if not getattr(self, 'buffering_handler', None):
            self.set_up_handlers()

        self.__announcement = None

        if announce_on_init:

            if init_announcement and isinstance(init_announcement, Announcement):
                self.__announcement = init_announcement
            elif init_announcement_template:
                self.__announcement = Announcement(
                        self,
                        init_announcement_template,
                        announcement_level
                        )

            if self.__announcement is None:
                self.__announcement = Announcement(
                        self,
                        Announcement.DEFAULT_INITIALIZATION_ANNOUNCEMENT,
                        announcement_level
                        )

            if auto_set_up:
                self.announce_initialization()

    @property
    def announcement(self) -> Announcement:
        return self.__announcement

    @announcement.setter
    @validate_type(Announcement, str, preferred_type=Announcement)
    def announcement(self, new: Union[Announcement, str]) -> None:
        if not self.announcement_made and isinstance(new, Announcement):
            self.__announcement = new

    @property
    def announcement_level(self) -> int:
        if not self.announcement:
            return self.console_level
        return self.announcement.announcement_level

    @property
    def announcement_made(self) -> bool:
        return self.announcement.announced

    @property
    def call_counts(self) -> dict:
        """
        Property to get the call counts of decorated methods.

        Returns:
            dict:
                A dictionary containing the call counts of decorated methods.

        """
        return self.__call_counts

    @property
    def child_names(self):
        return self.get_child_names()

    @property
    def children(self) -> List[InspyLogger]:
        return self.__children

    @children.deleter
    def children(self):
        self.__children = []

    @property
    def console_level(self) -> int:
        """
        Returns the logging level for the console.

        Returns:
            int:
                The logging level for the console.
        """
        return self.__console_level

    @console_level.setter
    @validate_type(
            str, int,
            preferred_type=str,
            conversion_funcs={int: translate_to_logging_level_str}
            )
    def console_level(self, level):
        """
        Sets the logging level for the console.

        Parameters:
            level:
                The logging level for the console.

        Returns:

        """
        if level.upper() not in LEVELS:
            raise ValueError(f'Invalid logging level: {level}. Please provide a valid logging level; one of {LEVELS}')

        self.__console_level = translate_to_logging_level(level)

        self.__apply_level_change('console')

    @property
    def console_level_name(self):
        self.internal('Test message')
        return get_level_name(self.console_level)

    @property
    def device(self):
        """
        Returns the logger instance.

        Returns:
            Logger:
                The logger instance.
        """
        logger = self
        logger.start = self.start.__get__(logger)
        return logger

    @property
    def file_level(self) -> int:
        """
        Returns the logging level for the file.

        Returns:
            int:
                The logging level for the file.
        """
        return self.__file_level

    @file_level.setter
    @validate_type(
            str, int,
            preferred_type=str,
            conversion_funcs={int: translate_to_logging_level_str}
            )
    def file_level(self, level):
        """
        Sets the logging level for the file.

        Parameters:
            level: The logging level for the file.

        Returns:
            None
        """
        if level.upper() not in LEVELS:
            raise ValueError(f'Invalid logging level: {level}. Please provide a valid logging level; one of {LEVELS}')

        self.__file_level = translate_to_logging_level(level)

        self.__apply_level_change('file')

    @property
    def file_level_name(self) -> str:
        return get_level_name(self.file_level)

    @property
    def file_path(self) -> Path:
        """
        Returns the path to the log file.

        Returns:
            Path:
                The path to the log file.

        """
        return self._file_path

    @file_path.setter
    def file_path(self, new):
        self._file_path = new

    @property
    def is_interactive_session(self) -> bool:
        return hasattr(sys, 'ps1') and sys.ps1

    @property
    def name(self) -> str:
        """
        Returns the name of the logger instance.

        Returns:
            str:
                The name of the logger instance.
        """
        return self.logger.name

    @property
    def no_file_logging(self) -> bool:
        return self.__no_file_logging

    @no_file_logging.setter
    @validate_type(bool)
    def no_file_logging(self, new):
        self.__no_file_logging = new

    @property
    def time_started(self) -> float:
        """
        Returns the time the logger was started.
        """
        return self.__time_started

    @property
    def warnings_issued(self) -> set:
        """
        Returns the one-time warnings issued by the logger.

        Since:
            v3.2.0

        Returns:
            set:
                A set containing the one-time warnings issued by the logger.
        """
        return self.__warnings_issued

    def __apply_level_change(self, handler_type):
        """
        Applies the level change to the specified handler type and the logger's children.

        Parameters:
            handler_type (str):
                The type of handler to apply the level change to.

        Returns:
            None

        Raises:
            ValueError:
                If the handler type is invalid.
        """
        handler_type = handler_type.lower()
        if handler_type not in HANDLER_TYPES:
            raise ValueError(
                    f'Invalid handler type: {handler_type}. '
                    f'Please provide a valid handler type; one of {HANDLER_TYPES}'
                    )

        level = getattr(self, f'{handler_type}_level')

        for handler in self.logger.handlers:
            if isinstance(handler, HANDLER_TYPES[handler_type]):
                handler.setLevel(level)

        self.logger.setLevel(translate_to_logging_level(level))

        for child in self.children:
            child.set_level(**{f'{handler_type}_level': level})

    def __build_name_from_caller(self, caller: inspect.FrameInfo, name: str = None):
        """
        Builds a name for a child logger from the caller's frame.

        Parameters:
            caller (inspect.FrameInfo):
                The frame of the caller.

        Returns:
            str:
                The name of the child-logger.

        """
        if name is None:
            name = caller.function

        caller_self = caller.frame.f_locals.get("self", None)
        separator = ":" if caller_self and hasattr(caller_self, name) else "."
        return f"{self.logger.name}{separator}{name}"

    def ensure_log_file_path(self):
        """
        Ensures that the log file path exists.
        """
        if not self.no_file_logging and not self.file_path.exists():
            self.file_path.parent.mkdir(parents=True, exist_ok=True)

            self.file_path.touch()

    def get_file_handler(self):
        """
        Fetches the file-handler for the logger.

        Returns:
            logging.FileHandler:
                The file handler for the logger.
        """
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler):
                return handler

    def has_child(self, name):
        """
        Checks if the logger has a child with the specified name.

        Parameters:
            name (str):
                The name of the child logger.

        Returns:
            bool:
                True if the logger has a child with the specified name, else False.
        """

        return name in self.child_names

    def isEnabledFor(self, level):
        return self.logger.isEnabledFor(level)

    @validate_type(str, Path, preferred_type=Path)
    def set_file_path(self, file_path):
        """
        Sets the file path for the logger.

        Parameters:
            file_path (str):
                The path to the log file.
        """
        old = self.file_path
        try:

            self.file_path = file_path
            self.ensure_log_file_path()
        except Exception:
            warn(f"Unable to set file path to {file_path}. Reverting to {old}")
            self.file_path = old
            raise

    def set_up_console(self):
        """
        Configures and attaches a console handler to the logger.
        """

        self.internal("Setting up console handler")
        console_handler = RichHandler(
                show_level=True, markup=True, rich_tracebacks=True,
                tracebacks_show_locals=True
                )
        formatter = CustomFormatter(
                f"[{self.logger.name}] %(message)s"
                )
        console_handler.setFormatter(formatter)
        console_handler.setLevel(self.__console_level)
        self.logger.addHandler(console_handler)

    def set_up_file(self):
        """
        Configures and attaches a file handler to the logger.
        """

        self.ensure_log_file_path()
        file_handler = logging.FileHandler(self.file_path)
        file_handler.setLevel(self.__file_level)
        formatter = CustomFormatter(
                "%(asctime)s - [%(name)s] - %(levelname)s - %(message)s |-| %(file_name)s:%(lineno)d"
                )
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

    def set_level(self, console_level=None, file_level=None, override=False, call_from_setter=False) -> None:
        """
        Updates the logging levels for both console and file handlers.

        Parameters:
            console_level:
                The logging level for the console.

            file_level:
                The logging level for the file.

            override (bool):
                Whether to override the `no_file_logging` option. Defaults to `False`.

            call_from_setter (bool):
                Whether the method was called from a setter method. Defaults to `False`.

        Returns:
            None
        """

        self.internal("Setting log levels")

        if not call_from_setter:

            # If we received a console level, update the console level.
            if console_level is not None:
                self.console_level = console_level

            # If we received a file level, update the file level.
            if file_level is not None:
                self.file_level = file_level

        if console_level is not None:
            self.__apply_level_change('console')

        if file_level is not None:
            self.__apply_level_change('file')

    @method_alias('add_child', 'add_child_logger', 'get_child_logger')
    def get_child(self, name=None, console_level=None, file_level=None, is_method=False, **kwargs) -> InspyLogger:
        """
        Retrieves or creates a nested child logger based on a dot-separated name.
    
        Parameters:
            name (str, optional):
                Dot-separated name representing the hierarchy of child loggers.
                E.g., 'child.method' will create or retrieve 'child' and then 'method' as nested children.
    
            console_level (int or str, optional):
                Console log level for the child logger(s).
    
            file_level (int or str, optional):
                File log level for the child logger(s).
    
        Returns:
            InspyLogger:
                The deepest child logger in the hierarchy specified by the name.
        """
        if name is None:
            # Get the name from the caller's function if not provided
            caller_frame = inspect.stack()[1]
            caller_name = caller_frame.function

            if caller_name in ("<module>", "<lambda>", None):
                caller_name = 'unnamed'

            name = caller_name

        if 'override' in kwargs:
            kwargs.pop('override')

        name_parts = name.split('.')
        current_logger = self

        for i, part in enumerate(name_parts):
            if i == len(name_parts) - 1 and is_method:
                # Last part of the name is a method, so we need to create a child logger for it
                # with ':' as the separator instead of '.' to avoid conflicts with existing loggers, and to indicate
                # that it's a method logger
                cl_name = f'{current_logger.name}:{part}'
            else:
                cl_name = f"{current_logger.name}.{part}"

            if found_child := current_logger.find_child_by_name(
                cl_name, exact_match=True
            ):
                current_logger = found_child
            else:
                console_level = console_level or current_logger.console_level
                file_level = file_level or current_logger.file_level

                child_logger = Logger(
                    name= cl_name,
                    console_level=console_level,
                    file_level=file_level,
                    parent=current_logger,
                    **kwargs
                )

                current_logger.children.append(child_logger)
                current_logger = child_logger

        return current_logger

    @method_alias('get_children_names', 'get_child_loggers')
    def get_child_names(self) -> List:
        """
        Fetches the names of all child loggers associated with this logger instance.
        """

        self.internal("Getting child logger names")
        return [child.name for child in self.children]

    def get_parent(self) -> InspyLogger:
        """
        Fetches the parent logger associated with this logger instance.
        """

        self.internal("Getting parent logger")
        return self.parent

    def find_child_by_name(self, name: str, case_sensitive=True, exact_match=False) -> (List, InspyLogger):
        """
        Searches for a child logger by its name.

        Parameters:
            name (str):
                The name of the child logger to search for.

            case_sensitive (bool, optional):
                Whether the search should be case-sensitive. Defaults to True.

            exact_match (bool, optional):
                Whether the search should only return exact matches. Defaults to False.

        Returns:
            list or Logger: If exact_match is True, returns the Logger instance if found, else returns an empty list.
                            If exact_match is False, returns a list of Logger instances whose names contain the
                            search term.
        """
        self.internal(f'Searching for child with name: {name}')
        results = []

        if not case_sensitive:
            name = name.lower()

        for logger in self.children:
            logger_name = logger.name if case_sensitive else logger.name.lower()
            if exact_match and name == logger_name:
                return logger
            elif not exact_match and name in logger_name:
                results.append(logger)

        return results

    @count_invocations
    def debug(self, message, *args, stack_level=2, **kwargs):
        """
        Logs a debug message.

        Parameters:
            message (str): The message to log.

            stack_level (int, optional):
                The stacklevel to use when logging. Defaults to 3.
        """
        self._log(logging.DEBUG, message, args=args, stacklevel=stack_level, **kwargs)

    @count_invocations
    def info(self, message, *args, stack_level=2, **kwargs):
        """
        Logs an info message.

        Parameters:
            message (str):
                The message to log.

            stack_level (int, optional):
                The stack-level to use when logging. Defaults to 2.

        Returns:
            None
        """
        self._log(logging.INFO, message, args=args, stacklevel=stack_level, **kwargs)

    def internal(self, message, *args, stack_level=2, **kwargs):
        """
        Logs an internal message.

        Parameters:
            message (str):
                The message to log.

            stack_level (int, optional):
                The stack-level to use when logging. Defaults to 2.
        """
        if self.logger.isEnabledFor(INTERNAL):
            self._log(INTERNAL, message, args=args, stacklevel=stack_level, **kwargs)

    @count_invocations
    def warning(self, message, *args, stack_level=2, **kwargs):
        """
        Logs a warning message.

        Parameters:
            message (str):
                The message to log.

            stack_level (int, optional):
                The stack-level to use when logging. Defaults to 2.

        Returns:
            None

        """
        self._log(logging.WARNING, message, args=args, stacklevel=stack_level, **kwargs)

    @count_invocations
    def error(self, message, *args, stack_level=2, **kwargs):
        """
        Logs an error message.

        Parameters:
            message (str):
                The message to log.

            stack_level (int, optional):
                The stack-level to use when logging. Defaults to 2.

        Returns:
            None
        """
        self._log(logging.ERROR, message, args=(), stacklevel=2, **kwargs)

    def __repr__(self):
        name = self.name
        hex_id = hex(id(self))
        if self.parent is not None:
            parent_part = f' | Parent Logger: {self.parent.name} |'
            if self.children:
                parent_part += f' | Number of children: {len(self.children)} |'
        else:
            parent_part = f' | This is a root logger with {len(self.children)} children. '

        if parent_part.endswith('|'):
            parent_part = str(parent_part[:-2])

        return f'<Logger: {name} w/ levels {self.console_level_name}, {self.file_level_name} at {hex_id}{parent_part}>'

    def announce_initialization(
            self,
            message_level: str = None,
            do_replace_placeholders=True,
            ):
        """
        Announces the initialization of the logger.

        Parameters:
            message_level (str, optional):
                The logging level to use for the announcement. Defaults to 'debug'.

            do_replace_placeholders (bool, optional):
                Whether to replace placeholders in the announcement. Defaults to True.

        Returns:
            None
        """
        if self.announcement_made:
            return

        if do_replace_placeholders:
            self.announcement.replace_placeholders()

        if message_level is None:
            message_level = self.announcement_level

        self.announcement.announcement_level = message_level

        self.announcement.announce()

    @classmethod
    def create_logger_for_caller(cls):
        """
        Creates a logger for the module that calls this method.

        Returns:
            Logger: An instance of the Logger class for the calling module.
        """
        if 'ipkernel' in sys.modules or 'IPython' in sys.modules:
            # We're running in an interactive environment, return a logger named 'interactive'

            return cls('Interactive-Python')
        frame = inspect.currentframe().f_back
        if module_path := cls._determine_module_path(frame):
            return cls(module_path)
        raise ValueError("Unable to determine module path for logger creation.")

    def replay_and_setup_handlers(self):
        """
        Replays the buffered logs and sets up the handlers for the logger.
        """
        if self.buffering_handler:
            self.buffering_handler.replay_logs(self.logger)

            # Remove the buffer handler
            self.logger.removeHandler(self.buffering_handler)

        # Set up the handlers
        if not self.logger.handlers:
            self.set_up_handlers()

    def set_up_handlers(self) -> None:
        """
        Sets up the handlers for the logger.
        """
        self.set_up_console()
        self.set_up_file()

    def to_dict(self):
        """
        Converts the logger properties into a dictionary format.

        Returns:
            dict:
                A dictionary containing the logger properties.

        """
        return {
                'Name':              self.name,
                'Console Level':     self.console_level_name,
                'File Level':        self.file_level_name,
                'Parent':            self.parent.name if self.parent else 'None',
                'Children':          {
                        'Count':   len(self.children),
                        'Names':   self.child_names,
                        'Loggers': [child.to_dict() for child in self.children]
                        },
                'Handlers':          {
                        'Count':    len(self.logger.handlers),
                        'Handlers': self.logger.handlers
                        },
                'Call Counts':       self.call_counts,
                'Buffering Handler': 'Yes' if getattr(self, 'buffering_handler', None) else 'No'
                }

    def start(self):
        """
        Starts the logger.
        """
        self.warning('InspyLogger.start() is deprecated.')
        if not self.logger.handlers:
            self.set_up_handlers()

        if hasattr(self, 'buffering_handler'):
            self.replay_and_setup_handlers()

        return self

    def warn_once(self, message, **kwargs):
        """
        Logs a warning message only once.

        Parameters:
            message:
                The warning message to be logged.

        Returns:
            None
        """
        if message not in self.__warnings_issued:
            self.warning(message, stack_level=kwargs.get('stack_level', 2))
            self.warnings_issued.add(message)

    @staticmethod
    def _determine_module_path(frame):
        """
        Determines the in-project path of the module from the call frame.

        Parameters:
            frame:
                The frame from which to determine the module path.

        Returns:
            str:
                The in-project path of the module.
        """
        if module := inspect.getmodule(frame):
            base_path = os.path.dirname(os.path.abspath(module.__file__))
            relative_path = os.path.relpath(frame.f_code.co_filename, base_path)
            return relative_path.replace(os.path.sep, '.').rstrip('.py')
        return None

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1):
        """
        Low-level logging implementation, passing stacklevel to findCaller.
        """
        if INTERACTIVE_SESSION:
            stacklevel -= 1

        if self.logger.isEnabledFor(level):
            self.logger._log(level, msg, args, exc_info, extra, stack_info, stacklevel + 1)

    def __rich__(self):
        # Create a rich table with logger properties
        from rich.table import Table
        from rich import box

        table = Table(title=f'[bold]Logger: {self.name}[/bold]', box=box.ASCII, padding=(0, 1, 0, 1))
        table.add_column('Property', justify='right', style='cyan', no_wrap=True)
        table.add_column('Value', justify='left', style='magenta', no_wrap=True)

        table.add_row('Name', self.name)
        table.add_row('Console Level', self.console_level_name)
        table.add_row('File Level', self.file_level_name)
        table.add_row('Parent', self.parent.name if self.parent else 'None')
        table.add_row('Children', str(len(self.children)))
        table.add_row('Handlers', str(len(self.logger.handlers)))

        if call_counts_str := ', '.join(
                f'{method}: {count}' for method, count in self.call_counts.items()
                ):
            table.add_row('Call Counts', call_counts_str)
        else:
            table.add_row('Call Counts', 'No method calls recorded')

        if getattr(self, 'buffering_handler', None):
            table.add_row('Buffering Handler', 'Yes' if self.buffering_handler else 'No')

        return table


def get_loggers():
    import gc

    gc.collect()
    loggers = [obj for obj in gc.get_objects() if isinstance(obj, InspyLogger)]

    return {str(logger).split(": ")[1].split(" w/")[0]: logger for logger in loggers}



import re
from typing import Optional
from inspy_logger.common import InspyLogger
from inspy_logger.models.announcement.placeholder import PlaceHolderString, PlaceHolderCollection


DEFAULT_DELIMITERS = {
        'open':  '{',
        'close': '}'

        }


def generate_placeholder_collection(
        template: str,
        owner: InspyLogger,
        open_delimiter: Optional[str] = DEFAULT_DELIMITERS['open'],
        close_delimiter: Optional[str] = DEFAULT_DELIMITERS['close']
        ) -> PlaceHolderCollection:
    """
    Generates a PlaceholderCollection based on a template string.

    Parameters:
        template (str):
            The template string containing placeholders.

        owner (Logger):
            The logger instance that owns the placeholders.

        open_delimiter (str):
            The opening delimiter for the placeholders. Defaults to '{'.

        close_delimiter (str):
            The closing delimiter for the placeholders. Defaults to '}'.

    Returns:
        PlaceHolderCollection:
            The generated :class:`PlaceholderCollection`.
    """
    placeholders = re.findall(rf'{re.escape(open_delimiter)}(.*?){re.escape(close_delimiter)}', template)
    return PlaceHolderCollection(
            owner = owner,
            template = template,
            open_delimiter=open_delimiter,
            close_delimiter=close_delimiter
            )



"""


Author:
    Inspyre Softworks

Project:
    inSPy-Logger

File:
    inspy_logger/types.py


Description:
    This module contains the LogLevel class for the logger.

    This class is used to define the log levels for the logger.

    The LogLevel class is a subclass of LogLevel

"""
from inspy_logger.constants import LEVEL_MAP
from inspy_logger.helpers import translate_to_logging_level
from inspy_logger.helpers.decorators import validate_type


class LogLevel:
    """
    A class for defining log levels for the logger.
    """
    LOG_LEVELS = {
        'DEBUG',
        'INFO',
        'WARNING',
        'ERROR',
        'CRITICAL',
    }

    __allowed_values = LOG_LEVELS

    __case_sensitive = None

    def __init__(self, log_level: str = 'INFO', case_sensitive=False):
        self.__level_name = None

        self.case_sensitive = case_sensitive

        levels = list(self.LOG_LEVELS)

        if isinstance(log_level, str):
            if not case_sensitive:
                for level in levels:
                    self.__allowed_values.add(level.lower())

        elif isinstance(log_level, int):
            self.level_name = self.from_int(log_level)

        self.level_name = log_level

    @property
    def allowed_values(self):
        return self.__allowed_values

    @property
    def case_sensitive(self):
        return self.__case_sensitive

    @case_sensitive.setter
    @validate_type(bool)
    def case_sensitive(self, case_sensitive: bool):
        self.__case_sensitive = case_sensitive

    @property
    def level_name(self):
        return self.__level_name

    @level_name.setter
    @validate_type(str, case_sensitive=False, allowed_values=LOG_LEVELS)
    def level_name(self, new_level: str):
        self.__level_name = new_level.upper()

    @classmethod
    def from_str(cls, level_string):
        level = level_string.upper()
        if level not in cls.LOG_LEVELS:
            raise ValueError(f"Invalid log level: {level_string}, must be one of {cls.LOG_LEVELS}")

        return cls(level)

    @classmethod
    def from_int(cls, level_integer):
        for level in cls.LOG_LEVELS:
            if translate_to_logging_level(level) == level_integer:
                return cls(level)

        raise ValueError(f"Invalid log level: {level_integer}, must be one of {cls.LOG_LEVELS}")

    def as_int(self):
        return translate_to_logging_level(self.level_name)

    def upper(self):
        return self.level_name.upper()

    def lower(self):
        return self.level_name.lower()

    def __str__(self):
        return self.level_name

    def __int__(self):
        return translate_to_logging_level(self.level_name)

    @classmethod
    def get_levels(cls):
        return [cls.DEBUG, cls.INFO, cls.WARNING, cls.ERROR, cls.CRITICAL]


class LogLevel(LogLevel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __convert_if_needed(self, other):
        if not isinstance(other, LogLevel):
            if isinstance(other, str):
                other = self.from_str(other)
            elif isinstance(other, int):
                other = self.from_int(other)

        return other

    def __eq__(self, other):
        other = self.__convert_if_needed(other)
        return self.level_name == other.level_name

    def __hash__(self):
        return hash(self.level_name)

    def __le__(self, other):
        other = self.__convert_if_needed(other)
        return self.as_int() <= other.as_int()


    def __lt__(self, other):
        other = self.__convert_if_needed(other)

        return self.as_int() < other.as_int()


    def __ge__(self, other):
        other = self.__convert_if_needed(other)
        return self.as_int() >= other.as_int()


    def __gt__(self, other):
        other = self.__convert_if_needed(other)
        return self.as_int() > other.as_int()


    def __repr__(self):
        return f"LogLevel: {self.level_name} | {self.as_int()} at {hex(id(self))})"


def assemble_log_levels():
    _ = list(LogLevel.LOG_LEVELS)
    levels = [level for level in _ if level.isupper()]

    levels = sorted(levels, key=lambda x: LEVEL_MAP[x.lower()])

    return {LogLevel(level) for level in levels}


if 'LOG_LEVELS' not in globals():
    LOG_LEVELS = assemble_log_levels()



"""


Author:
    Inspyre Softworks

Project:
    inSPy-Logger

File:
    inspy_logger/helpers/decorators.py


Description:
    Contains decorators used by the inSPy-Logger package.

"""
from functools import wraps
from typing import Callable, Optional


__all__ = [
    'method_alias',
    'method_logger',
    'add_aliases',
    'count_invocations',
    'validate_type'
]


def property_logging(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.creating_logger:
            return
        if found := self.class_logger.find_child_by_name(func.__name__, exact_match=True):
            log = found
        else:
            self.creating_logger = True
            log = self.class_logger.add_child(func.__name__)
            self.creating_logger = False

        log.debug(f'Getting {func.__name__} from instance of {self.__class__.__name__}...', stack_level=4)
        res = func(self, *args, **kwargs)
        log.debug(f'Got {func.__name__} from instance of {self.__class__.__name__}: {res}', stack_level=4)

        return res
    return wrapper



def method_alias(*alias_names: (str, list[str])):
    """
    A decorator that allows you to add aliases to a class's method.

    Parameters:
        *alias_names (str, list[str]):
            The name(s) of the alias(es) to add.

    Returns:
        method:
            The decorated method.
    """
    def method_decorator(meth):
        meth._alias_names = alias_names
        return meth

    return method_decorator


def method_logger(func):
    """
    A decorator to log a method's execution.

    Parameters:
        func (method):
            The method to log.

    Returns:
        method:
            The decorated method.
    """
    def wrapper(self, *args, **kwargs):
        # Set up the logger with method-specific config.
        name = f'{self.__class__.__name__}.{func.__name__}'
        logger = self.class_logger.get_child()

        # Add the logger to the instance for use in the method
        self._current_logger = logger

        # Execute the called method
        res = func(self, *args, **kwargs)

        # Remove the logger after method execution
        del self._current_logger

        return res
    return wrapper


def add_aliases(cls):
    """
    A decorator to add aliases to a class's methods.

    Parameters:
        cls (class):
            The class to add aliases to.

    Returns:
        class:
            The decorated class.
    """
    for name, method in list(cls.__dict__.items()):
        if hasattr(method, '_alias_names'):
            for alias_name in method._alias_names:
                setattr(cls, alias_name, method)
    return cls


def count_invocations(method):
    """
    Decorator that counts the number of invocations of the method it decorates.

    Parameters:
        method (Callable):
            The method to count invocations for.

    Returns:
        Callable:
            The decorated method.
    """
    def wrapper(self, *args, **kwargs):
        method_name = method.__name__
        if method_name not in self.call_counts:
            self.call_counts[method_name] = 0

        self.call_counts[method_name] += 1

        return method(self, *args, **kwargs)

    return wrapper


def validate_type(
        *allowed_types,
        preferred_type=None,
        allowed_values=None,
        case_sensitive=True,
        conversion_funcs: Optional[dict] = None):
    """
    A decorator for validating the type and optionally the value of a value passed to a class
    property setter, with an option to convert to a preferred type if specified and to enforce
    value restrictions.

    Parameters:
        preferred_type (type, optional):
            The preferred type to which values should be converted if possible. If None, no
            conversion is attempted.

        *allowed_types:
            Variable length list of allowed types for the property value.

        allowed_values (iterable, optional):
            An iterable of values that are allowed. If None, all values of the correct type are allowed.

        case_sensitive (bool, optional):
            Specifies whether string comparisons should be case-sensitive. Defaults to True. Ignored for non-string
            types.

        conversion_funcs (dict, optional):
            A dictionary of conversion functions to use for converting

    Returns:
        A decorator function for the property setter.

    Raises:
        TypeError:
            If the incoming value does not match one of the allowed types, cannot be onverted to the preferred type,
            or is not in the list of allowed values.

        ValueError:
            If the value is of the correct type but not in the allowed values list.

    Example:
        >>> class MyClass:
        ...     @property
        ...     def my_property(self):
        ...         return self._my_property
        ...
        ...     @my_property.setter
        ...     @validate_type(str, allowed_values=["hello", "world"], case_sensitive=False)
        ...     def my_property(self, value):
        ...         self._my_property = value
        ...
        >>> obj = MyClass()
        >>> obj.my_property = "HELLO"  # This is accepted and set
        >>> obj.my_property = "hello"  # This is also accepted and set
        >>> obj.my_property = "Hi"  # This raises ValueError
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, value):
            if not isinstance(value, tuple(allowed_types)):
                allowed = ', '.join([t.__name__ for t in allowed_types])
                raise TypeError(f"Value must be of type {allowed}, got type {type(value).__name__}")

            if preferred_type and not isinstance(value, preferred_type):
                try:
                    if conversion_funcs and type(value) in conversion_funcs:
                        value = conversion_funcs[type(value)](value)
                    else:
                        value = preferred_type(value)
                except Exception as e:
                    raise TypeError(f"Could not convert value to preferred type {preferred_type.__name__}: {e}")

            if allowed_values is not None:
                if isinstance(value, str) and not case_sensitive:
                    validation_passed = any(val.lower() == value.lower() for val in allowed_values)
                else:
                    validation_passed = value in allowed_values

                if not validation_passed:
                    allowed_vals = ', '.join(map(str, allowed_values))
                    raise ValueError(f"Value must be one of {allowed_vals}, got {value}")

            return func(self, value)

        return wrapper

    return decorator


def validate_func_args(*validators):
    """
    A decorator for validating the type and optionally the value of arguments passed to a function.
    Each argument can have its own set of allowed types, a preferred type, and allowed values.

    Parameters:
        *validators (dict): Each dictionary in validators specifies the validation rules for one argument.
                            The keys in the dictionary can be:
                            - 'allowed_types': A tuple of allowed types for the argument.
                            - 'preferred_type': The preferred type to which the argument should be converted if possible.
                            - 'allowed_values': A list of values that the argument is allowed to have.
                            - 'case_sensitive': Specifies whether string comparisons should be case sensitive.

    Returns:
        A decorator function for the function.

    Raises:
        TypeError: If an argument does not match one of the allowed types or cannot be converted to the preferred type.
        ValueError: If the argument is of the correct type but not in the allowed values list.

    Example:
        >>> @validate_func_args(
        ...     {'allowed_types': (str,), 'allowed_values': ["hello", "world"], 'case_sensitive': False},
        ...     {'allowed_types': (int,), 'preferred_type': float}
        ... )
        ... def my_function(word, number):
        ...     print(f"{word} {number}")
        ...
        >>> my_function("HELLO", 123)  # This is valid and prints 'hello 123.0'
        >>> my_function("unknown", 123)  # This raises ValueError for the first argument
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            new_args = []
            for i, (arg, validator) in enumerate(zip(args, validators)):
                allowed_types = validator.get('allowed_types', ())
                preferred_type = validator.get('preferred_type')
                allowed_values = validator.get('allowed_values')
                case_sensitive = validator.get('case_sensitive', True)

                if not isinstance(arg, allowed_types):
                    allowed = ', '.join([t.__name__ for t in allowed_types])
                    raise TypeError(f"Argument {i} must be of type {allowed}, got type {type(arg).__name__}")

                if preferred_type and not isinstance(arg, preferred_type):
                    try:
                        arg = preferred_type(arg)
                    except Exception as e:
                        raise TypeError(
                            f"Could not convert argument {i} to preferred type {preferred_type.__name__}: {e}")

                if allowed_values is not None:
                    if isinstance(arg, str) and not case_sensitive:
                        validation_passed = any(val.lower() == arg.lower() for val in allowed_values)
                    else:
                        validation_passed = arg in allowed_values

                    if not validation_passed:
                        allowed_vals = ', '.join(map(str, allowed_values))
                        raise ValueError(f"Argument {i} must be one of {allowed_vals}, got {arg}")

                new_args.append(arg)

            return func(*new_args, **kwargs)

        return wrapper

    return decorator



"""
Author: Taylor B. tayjaybabee@gmail.com
Date: 2024-11-23 13:07:52
LastEditors: Taylor B. tayjaybabee@gmail.com
LastEditTime: 2024-11-27 20:23:34
FilePath: inspy_logger/engine/adapters/__init__.py
Description: 这是默认设置,可以在设置》工具》File Description中进行配置
"""
from inspy_logger.engine.adapters.thread import ThreadAdapter



#!/usr/bin/env python3
import contextlib
import sys
import inspect
import os
import logging
from pypattyrn.behavioral.null import Null
from inspy_logger.common import PROG_NAME as ISL_PROG_NAME, DEFAULT_LOGGING_LEVEL, DEFAULT_LOG_FILE_PATH, LEVELS
from inspy_logger.helpers import find_variable_in_call_stack, check_preemptive_level_set, find_argument_parser, determine_start_block, determine_level

from inspy_logger.helpers import (
    translate_to_logging_level,
    clean_module_name,
    CustomFormatter,
    determine_client_prog_name,
    determine_log_file_path
)

# Existing log record factory
old_factory = logging.getLogRecordFactory()


def record_factory(*args, **kwargs):
    record = old_factory(*args, **kwargs)
    # Set the file_name attribute to the name of the file where the log is called
    frame = inspect.stack()[1]
    record.file_name = frame.filename
    return record


logging.setLogRecordFactory(record_factory)


from inspy_logger.engine import Logger
from inspy_logger.helpers import get_existing_logger

__all__ = [
    "clean_module_name",
    "CustomFormatter",
    "DEFAULT_LOGGING_LEVEL",
    "determine_level",
    "InspyLogger",
    "LOG_DEVICE",
    "Logger",
    "ISL_PROG_NAME",
    "translate_to_logging_level",
]

BLOCKED = determine_start_block()

LOG_FILE = determine_log_file_path()

LOG_DEVICE = None if BLOCKED else Logger(ISL_PROG_NAME)

MODULE_OBJ = sys.modules[__name__]

CLIENT_PROG_NAME = determine_client_prog_name()

INIT_LOG_LEVEL = determine_level(CLIENT_PROG_NAME)

INTERACTIVE_SESSION = find_variable_in_call_stack('INSPY_INTERACTIVE_SESSION', default=False)


def start_logger(override_block=True):
    """
    Starts the logger.

    Parameters:
        override_block (Union[bool, None], optional):
            A flag to override the block on the logger. Defaults to True.

    Note:
        You only need to run this function if you have blocked the logger from starting automatically.

    Returns:

    """
    global LOG_DEVICE
    global CLIENT_PROG_NAME
    global INIT_LOG_LEVEL

    if not BLOCKED or override_block:
        LOG_DEVICE = Logger(ISL_PROG_NAME)
        LOG_DEVICE.set_level(INIT_LOG_LEVEL)

        if CLIENT_PROG_NAME:

            prog_logger = Logger(
                CLIENT_PROG_NAME,
                LOG_DEVICE.console_level,
                LOG_DEVICE.file_level,
                file_name=DEFAULT_LOG_FILE_PATH.name,
                file_path=DEFAULT_LOG_FILE_PATH.parent
            )
            __all__.append('PROG_LOGGER')

        LOG_DEVICE.replay_and_setup_handlers()
    elif CLIENT_PROG_NAME:
        prog_logger = Null()

    with contextlib.suppress(NameError):
        global PROG_LOGGER

        PROG_LOGGER = prog_logger

        if isinstance(PROG_LOGGER, Null):
            from rich import print
            print("The logger has been blocked from starting. To start the logger, run `start_logger()`.")


InspyLogger = Logger


def _get_loggable():
    from inspy_logger.helpers.base_classes import Loggable
    import bisect
    # Add the `Loggable` class to `__all__` in alphabetical order.
    bisect.insort(__all__, 'Loggable')

    return Loggable


Loggable = _get_loggable()


start_logger(override_block=False)


deprecation_warning_deliveries = 0


def getLogger(*args, **kwargs):
    global deprecation_warning_deliveries

    if deprecation_warning_deliveries < 5:
        LOG_DEVICE.warning('The `getLogger` function is deprecated. Use `InspyLogger` instead.')
        deprecation_warning_deliveries += 1

    return InspyLogger(*args, **kwargs)



import inspect
import logging
import re
from logging import Formatter
from pathlib import Path

from inspy_logger.__about__ import __PROG__
from inspy_logger.constants import LEVEL_MAP, DEFAULT_LOGGING_LEVEL, HANDLER_TYPES
from inspy_logger.helpers.decorators import validate_type
from inspy_logger.helpers.descriptors import RestrictedSetter
from typing import Any, Optional, Union

"""
This module contains utility functions and classes for handling logging and
performing certain string and number operations.
"""


__all__ = [
    "check_preemptive_level_set",
    "clean_module_name",
    "CustomFormatter",
    "determine_client_prog_name",
    "determine_start_block",
    "find_argument_parser",
    "find_key_by_value",
    "find_valid_vars_in_call_stack",
    "find_variable_in_call_stack",
    "get_level_name",
    "is_number",
    "translate_to_logging_level",
    "translate_to_logging_level_str",
    "VALID_DEV_NAME_VARS",
    "VALID_LOG_FILE_VARS",
    "VALID_PROG_NAME_VARS",
    'determine_log_file_path',
    'determine_log_filepath',
    'RestrictedSetter',
]


class ValidVars:
    VALID_MODES = ['all', 'strict']
    _mode = RestrictedSetter(
        'mode',
        allowed_types=str,
        allowed_values=VALID_MODES,
        initial='strict'
    )

    def __init__(self, general_vars, strict_vars, mode=None):
        self.mode = mode or 'strict'
        self.__general_vars = None
        self.__strict_vars = None

        self.general_vars = general_vars
        self.strict_vars = strict_vars

    @property
    def general_vars(self):
        return self.__general_vars

    @general_vars.setter
    @validate_type(str, list, tuple, preferred_type=tuple)
    def general_vars(self, new):
        self.__general_vars = new

    @property
    def strict_vars(self):
        return self.__strict_vars

    @strict_vars.setter
    @validate_type(str, list, tuple, preferred_type=tuple)
    def strict_vars(self, new):
        self.__strict_vars = new

    @property
    def valid_vars(self):
        if self.mode == 'all':
            return self.general_vars + self.strict_vars
        elif self.mode == 'strict':
            return self.strict_vars

    def __repr__(self):
        return f"{self.__class__.__name__}(mode={self.mode}, "\
               f"general_vars={self.general_vars}, strict_vars={self.strict_vars})"

    def __str__(self):
        return self.__rich__()

    def __rich__(self):
        from rich.table import Table
        from rich.console import Console

        table = Table(
            title=self.__class__.__name__,
            show_lines=True,
            header_style='bold cyan',
            row_styles=['bold', 'italic'],
            highlight=True

        )
        table.add_column("Attribute")
        table.add_column("Value")

        table.add_row("Mode", self.mode)
        table.add_row("General Vars", ', '.join(self.general_vars))
        table.add_row("Strict Vars", ', '.join(self.strict_vars))

        console = Console()
        console.print(table)


def replace_placeholders(
        string: str,
        replacement_map: dict,
        ignore_case: bool = False,
        open_brace: str = '{',
        close_brace: str ='}'

    ):
    """
    Replaces placeholders in a string with values from a replacement map.

    Parameters:
        string (str):
            The string to replace placeholders in.

        replacement_map (dict):
            The map of placeholders and their replacements.

        ignore_case (bool, optional):
            A flag indicating whether to ignore the case of the placeholders. Defaults to False.

        open_brace (str, optional):
            The opening brace of the placeholders. Defaults to '{'.

        close_brace (str, optional):
            The closing brace of the placeholders. Defaults to '}'.

    Returns:
        str:
            The string with placeholders replaced.

    Example:
        >>> replace_placeholders("Hello, {name}!", {"<name>": "John"})
        "Hello, John!"

    Since:
        v3.1.0
    """
    for placeholder, replacement in replacement_map.items():
        full_placeholder = f'{open_brace}{placeholder}{close_brace}'
        string = string.replace(full_placeholder, replacement)

    return string



VALID_LOG_FILE_VARS = [
    'LOG_FILE',
    'LOG_FILE_PATH',
    'INSPY_LOG_FILE',
    'INSPY_LOG_PATH',
    'INSPY_LOG_FILEPATH',
    'INSPY_LOG_FILE_PATH',
    'INSPY_LOGGER_FILEPATH',
    'INSPY_LOGGER_PATH',
    'INSPY_LOGGER_FILE_PATH'
]


VALID_LOG_LEVEL_VARS = [
    'LOG_LEVEL',
    'LOGGING_LEVEL',
    'INSPY_LOG_LEVEL',
    'INSPY_LOGGING_LEVEL',
]


VALID_ARGS_VARS = [
    'ARGUMENT_PARSER',
    'ARGS_PARSER',
    'ARGS',
    'PARSED_ARGS',
]

VALID_PROG_NAME_VARS = [
    'PROG',
    'PROG_NAME',
    'PROGRAM_NAME',
    'PROGRAM',
    'PROGNAME',
    '__PROG__',
]

VALID_DEV_NAME_VARS = ValidVars(
    [
        'DEV_NAME',
        'DEVNAME',
        'DEVELOPER_NAME',
        'DEVELOPER',
        'DEV',
        'AUTHOR',
        'SOFTWARE_ORG'
    ],
    [
        'INSPY_LOG_DEV_NAME',
        'INSPY_DEV_NAME',
        'INSPY_LOG_AUTHOR',
        'INSPY_AUTHOR',
        'INSPY_SOFTWARE_ORG',
        'INSPY_LOG_SOFTWARE_ORG'
    ],
    mode='strict'
)

class CustomFormatter(Formatter):
    """
    CustomFormatter extends the logging.Formatter class to provide a custom
    formatting behavior. Specifically, it replaces '<ipython-input-...>'
    patterns in record.pathname with 'iPython'.
    """

    def format(self, record):
        """
        Replaces <ipython-input-...> pattern in record.pathname with 'iPython'.

        Parameters:
            record (logging.LogRecord): The record to format.

        Returns:
            str: The formatted record.
        """
        # Replace <ipython-input-...> pattern in record.pathname
        record.pathname = re.sub(
            r"<ipython-input-\d+-\w+>|<module>", "iPython", record.pathname
        )
        return super().format(record)


def clean_module_name(module_name):
    """
    Replaces <ipython-input-...> pattern in the given module name with 'iPython'.

    Parameters:
        module_name (str): The module name to clean.

    Returns:
        str: The cleaned module name.
    """
    return re.sub(r"<ipython-input-\d+-\w+>", "iPython", module_name)


def determine_level(client_prog_name):
    """
    Determines the level at which to output logs to the console.

    Returns:
        int:
            The level at which to output logs to the console.
    """

    level = DEFAULT_LOGGING_LEVEL

    if _preemptive_set := check_preemptive_level_set():
        level = translate_to_logging_level(_preemptive_set)

    if client_prog_name:
        if arg_parser := find_argument_parser():
            from inspy_logger.helpers.command_line import add_argument
            add_argument(arg_parser, level)
            args = arg_parser.parse_args()

            level = translate_to_logging_level(args.log_level)

    return level


def find_key_by_value(dictionary: dict, value: Any) -> Union[str, None]:
    """
    Find the first key in the :param:`dictionary` that corresponds to the given :param:`value`.

    Parameters:
        dictionary (dict):
            The dictionary to search.

        value (Any):
            The value to search for.

    Returns:
        key (str):
            The first key that matches the value. If no key is found, return None.

    Example:
        >>> d = {'a': 1, 'b': 2, 'c': 3}
        >>> find_key_by_value(d, 2)  # Returns 'b'
    """
    for key, val in dictionary.items():
        if val == value:
            return key

    return None



def is_number(string, force_integer=False, rounding=None):
    """
    Checks if a given string can be converted to a number and optionally
    rounds or converts the result to an integer.

    Parameters:
        string (str): The string to check.
        force_integer (bool, optional): If True, the result will be converted to an integer.
        rounding (int, optional): The number of decimal places to round to.

    Returns:
        float|int|str: The converted number, or the original string if it cannot be converted.
    """
    num = None

    #print(f"Received string {string}")

    if isinstance(string, (int, float)):
        print("Detected that received string is actually an integer or float...")
        num = string
        print(f'Num is now "{num}" after detecting that "string" is indeed a string.')
    elif isinstance(string, str):
        print("Detected that received string is indeed a string.")

        try:
            print("Attempting to convert the string to a float...")
            # Try to convert the string to a float.
            num = float(string)
            print(
                f"After conversion attempt from string to float, {num} is {type(num)}"
            )

            # If rounding is specified, round the number.
            if rounding is not None and isinstance(rounding, int) and rounding >= 0:
                print("Detected parameters to return the number rounded.")
                num = round(num, rounding)
                print(f"After rounding, the number is {num}")

        except ValueError as e:
            print(e)
            # If a ValueError is raised, the string is not a number.
            num = string

    if force_integer and not isinstance(num, str):
        print("Detecting that we were instructed to return an integer.")
        num = int(num)
        print(f"After converting the number to an integer it is now; {num}")

    print(f"Returning number which is {num}")
    return num


def translate_to_logging_level(level_str):
    """
    Translates a given string to a logging level.

    Parameters:
        level_str (str): The string to translate.

    Returns:
        int: The corresponding logging level, or None if the string does not correspond to a level.
    """
    if not isinstance(level_str, str):
        return level_str

    level_str = level_str.lower()

    # Return the ps_logging level if it exists, else return None
    return LEVEL_MAP.get(level_str)


def translate_to_logging_level_str(level):
    """
    Translates a given logging level to a string.

    Parameters:
        level (int):
            The logging level to translate.

    Returns:
        str:
            The corresponding logging level string, or None if the level does not correspond to a string.

    Raises:
        ValueError:
            If the logging level is invalid.

    Example:
        >>> translate_to_logging_level_str(logging.INFO)
        'INFO'
    """
    res = find_key_by_value(LEVEL_MAP, level)
    if not res:
        raise ValueError(f"Invalid logging level: {level}")

    return res.upper()


def get_level_name(level: int) -> (str, None):
    """
    Gets the name of the logging level.

    Parameters:
        level (int):
            The logging level.

    Returns:
        str:
            The name of the logging level.

    Raises:
        ValueError:
            If the logging level is invalid.
    """
    return translate_to_logging_level_str(level)


# def find_variable_in_call_stack(var_name, default=None):
#     """
#     Searches for a variable in the namespaces of all modules in the call stack.
#
#     Parameters:
#         var_name (str): The name of the variable to find.
#         default: Default value to return if the variable is not found.
#
#     Returns:
#         The first occurrence of the variable found in the call stack or the default value.
#     """
#     frame = inspect.currentframe()
#
#     try:
#         # Traverse the stack in the order of calling
#         while frame:
#             module = inspect.getmodule(frame)
#             if module and hasattr(module, var_name):
#                 print('FOUND THE VARIABLE IN THE CALL STACK!')
#                 return getattr(module, var_name)
#
#             frame = frame.f_back
#
#         return default
#     finally:
#         del frame


def find_variable_in_call_stack(var_name, ignore_inspy_logger: bool = False, default=None):
    """
    Searches for a variable in the namespaces of all modules in the call stack.

    Parameters:
        var_name (str): The name of the variable to find.
        ignore_inspy_logger (bool): Whether to ignore variables from the Inspy-Logger module.
        default: Default value to return if the variable is not found.

    Returns:
        The first occurrence of the variable found in the call stack or the default value.
    """
    frame = inspect.currentframe()

    try:
        while frame:
            caller_locals = frame.f_locals
            caller_globals = frame.f_globals
            caller_name = frame.f_back.f_code.co_name if frame.f_back else None

            for var_source in [caller_locals, caller_globals]:
                if var_name in var_source:
                    var_value = var_source[var_name]

                    # Get the module where the variable was defined
                    var_module = inspect.getmodule(frame)

                    if ignore_inspy_logger and (var_module and 'inspy_logger' in var_module.__name__):
                        continue
                    return var_value

            frame = frame.f_back  # Move to the previous frame after checking both namespaces
        return default  # Variable not found in the call stack
    finally:
        del frame


def iterate_valid_vars(valid_vars, mode='strict'):
    """
    Iterates through the valid variables based on the mode.

    Parameters:
        valid_vars (list): The list of valid variables to iterate through.
        mode (str, optional): The mode to use. Defaults to 'strict'.

    Yields:
        str: The next valid variable.
    """
    if mode == 'all':
        yield from valid_vars
    elif mode == 'strict':
        for var in valid_vars:
            if var.startswith('INSPY'):
                yield var


def find_valid_vars_in_call_stack(valid_vars, mode='strict'):
    """
    Finds the first valid variable in the call stack.

    Parameters:
        valid_vars (list):
            The list of valid variables to search for.

        mode (str, optional):
            The mode to use. Defaults to 'strict'.

    Returns:
        str: The first valid variable found in the call stack.
    """
    for var in iterate_valid_vars(valid_vars, mode):
        if var_value := find_variable_in_call_stack(var, ignore_inspy_logger=True):
            return var_value


def check_preemptive_level_set() -> (str, None):
    """
    Checks if the preemptive level has been set in the call stack. If not, it checks to see if an argument parser exists
    in the stack

    Returns:
        bool: True if the preemptive level has been set, False otherwise.
    """
    for var in VALID_LOG_LEVEL_VARS:
        if level := find_variable_in_call_stack(var):
            return level

    if arg_parser := find_argument_parser():
        if hasattr(arg_parser, 'parsed') and hasattr(arg_parser.parsed, 'log_level'):
            return arg_parser.parsed.log_level

    return None


def determine_client_prog_name() -> (str, None):
    """
    Determines the name of the program that is calling the logger.

    Returns:
        str:
            The name of the program that is calling the logger.

        None:
            If the name of the program that is calling the logger cannot be determined.
    """
    valid_vars = [
        'PROG',
        'PROG_NAME',
        'PROGRAM_NAME',
        'PROGRAM',
        'PROGNAME',
        '__PROG__'
    ]
    for var in valid_vars:
        if prog_name := find_variable_in_call_stack(var, ignore_inspy_logger=True):
            if prog_name != __PROG__:
                return prog_name


def determine_log_file_path() -> (Path, None):
    """
    Determines the path to the log file.

    This function first checks for the presence of environment variables that specify the path to the log file. If
    none are found, it then searches the call stack for variables that specify the path to the log file. If none are
    found, it then searches the call stack for an argument parser and uses the default log file path specified in the
    argument parser.

    Returns:
        str:
            The path to the log file.

        None:
            If the path to the log file cannot be determined.
    """

    for var in VALID_LOG_FILE_VARS:
        if log_file_path := find_variable_in_call_stack(var):
            return log_file_path

    if arg_parser := find_argument_parser():
        if hasattr(arg_parser, 'parsed') and hasattr(arg_parser.parsed, 'log_file'):
            return arg_parser.parsed.log_file

    return None


def determine_log_filepath() -> (Path, None):
    """
    Determines the path to the log file.

    Returns:
        str:
            The path to the log file.

        None:
            If the path to the log file cannot be determined.
    """
    return determine_log_file_path()


def determine_start_block() -> bool:
    """
    Determines if the logger should be blocked from starting.

    To set this, set the `BLOCK_LOGGER_START` variable in the calling module to `True` before importing Inspy-Logger.

    Returns:
        bool:
            True if the logger should be blocked from starting, False otherwise.

    """
    return find_variable_in_call_stack("BLOCK_LOGGER_START", False)


def find_argument_parser():
    """
    Finds the argument parser in the call stack.

    Returns:
        ArgumentParser:
            The argument parser in the call stack.

        None:
            If the argument parser cannot be found in the call stack.
    """
    for var in VALID_ARGS_VARS:
        if arg_parser := find_variable_in_call_stack(var):
            return arg_parser


def get_existing_logger(logger_name):
    from inspy_logger.engine import get_loggers
    loggers = get_loggers()

    if logger_name in loggers:
        return loggers[logger_name]



"""

File: inspy_logger/__about__.py
Project: inSPy-Logger
Description:
    Holds information about the inSPy-Logger program and its version.

Created: 11/3/22 - 21:05:06

Since:
    v2.1.1

"""
from inspy_logger.version import parse_version
from inspy_logger.version import parse_version

__PROG__ = 'inSPy-Logger'
__VERSION__: str = parse_version()
__AUTHORS__: list[tuple[str, str]] = [
        ('Inspyre-Softworks', 'https://inspyre.tech'),
        ('Taylor-Jayde Blackstone', '<t.blackstone@inspyre.tech>')
]

from inspy_logger.version import parse_version

__PROG__ = 'inSPy-Logger'
__VERSION__: str = parse_version()
__AUTHORS__: list[tuple[str, str]] = [
        ('Inspyre-Softworks', 'https://inspyre.tech'),
        ('Taylor-Jayde Blackstone', '<t.blackstone@inspyre.tech')
]

SOFTWARE_ORG = __AUTHORS__[0][0]



"""


Author: 
    Inspyre Softworks

Project:
    inSPy-Logger

File: 
    inspy_logger/engine/errors.py
 

Description:
    

"""
from inspy_logger.errors import InSPyLoggerError


__all__ = [
    'InSPyLoggerEngineError',
]


class InSPyLoggerEngineError(InSPyLoggerError):
    pass


class InSPyLoggerEngineConfigError(InSPyLoggerEngineError):
    pass


class InSPyLoggerLogFileError(InSPyLoggerEngineError):
    message = "[Log File Error]\n" \
              "An error occurred while trying to access the log file."
    prefix = "Some further information from the caller:\n"

    def __init__(self, message=None, skip_print=False):
        self.__child_caller_frame = inspect.stack()[1]
        self.__child_caller_name = self.__child_caller_frame[3]

        self.message = f"[{self.__class__.__name__}] - {self.message}"

        message = f"{self.prefix} - {message}" if message is not None else ''

        self.message += message

        super(InSPyLoggerLogFileError, self).__init__(message=self.message, skip_print=skip_print)



from argparse import ArgumentParser
from inspy_logger.common import DEFAULT_LOGGING_LEVEL, LEVELS


def add_argument(parser: ArgumentParser, default=DEFAULT_LOGGING_LEVEL):
    """
    Add an argument to the given ArgumentParser object.

    Parameters:
        parser (ArgumentParser): The ArgumentParser object to add the argument to.
        default (Any): The default value for the argument. Defaults to DEFAULT_LOGGING_LEVEL.

    Returns:
       None

    Raises:
        None

    Description:
        This function adds an argument to the given ArgumentParser object. The argument is used to set the level at which to output logs to the console. The argument can be set using the '-l' or '--log-level' flags. The default value for the argument is set to DEFAULT_LOGGING_LEVEL. The valid choices for the argument are defined in the LEVELS variable.
    """
    parser.add_argument(
        '-l', '--log-level',
        action='store',
        default=default,
        choices=LEVELS,
        help='Set the level at which to output logs to the console.'
    )



from typing import Optional, Union
from inspy_logger.common import InspyLogger, LEVEL_NAMES
from inspy_logger.helpers import translate_to_logging_level_str
from inspy_logger.helpers.decorators import validate_type
from inspy_logger.helpers.descriptors import RestrictedSetter
from inspy_logger.models.announcement.placeholder import PlaceHolderString, PlaceHolderCollection
from inspy_logger.models.announcement.placeholder.helpers import generate_placeholder_collection

from warnings import warn


class Announcement:

    DEFAULT_INITIALIZATION_ANNOUNCEMENT = 'Initialized {name}.'

    def __init__(self,
                 owner: InspyLogger,
                 template: Optional[str] = None,
                 announcement_level: Optional[Union[int, str]] = None
                 ):
        """
        Initialize the announcement.

        Parameters:
            owner (InspyLogger):
                The owner of the announcement.

            template (Optional[str]):
                The template for the announcement.

            announcement_level (Optional[Union[int, str]]):
                The logging level for the announcement.

        Returns:
            None

        Raises:
            ValueError:
                If the logging level is invalid.
        """

        self.__owner = None
        self.__template = None
        self.__replacement_map = None
        self.__placeholders = None
        self.__announced = False
        self.__announcement_text = None
        self.__announcement_level = None
        self.__replaced = False

        self.announcement_text = RestrictedSetter('announcement_text', allowed_types=str, restrict_setter=True)
        if announcement_level is not None:
            self.announcement_level = announcement_level
        else:
            self.announcement_level = 'debug'

        self.owner = owner
        self.template = template
        self.placeholders = generate_placeholder_collection(self.template, self.owner)

    @property
    def announced(self) -> bool:
        """
        The announcement status. True if the announcement has been announced, False otherwise.

        Returns:
            bool:
                The announcement status.
        """
        return self.__announced

    @announced.setter
    @validate_type(bool)
    def announced(self, new: bool) -> None:
        """
        Set the announcement status.

        Parameters:
            new (bool):
                The announcement status.

        Returns:
            None

        Raises:
            ValueError:
                If the announcement has already been announced.
        """
        if not self.announced and new:
            self.__announced = new

        else:
            warn('Announcement has already been announced. Property `announced` remains unchanged!')

    @property
    def announcement_level(self):
        """
        The logging level for the announcement.

        Returns:
            str:
                The logging level for the announcement.
        """
        return self.__announcement_level

    @announcement_level.setter
    @validate_type(int, str, None, preferred_type=str, conversion_funcs={int: translate_to_logging_level_str})
    def announcement_level(self, new: str):
        """
        Set the logging level for the announcement.

        Parameters:
            new (Union[str, int]):
                The logging level for the announcement.

        Returns:
            None

        Raises:
            ValueError:
                If the logging level is invalid.
        """
        if self.announced:
            warn('Announcement has already been announced. Property `announcement_level` remains unchanged!')

        elif new.upper() not in LEVEL_NAMES:
            raise ValueError(f'Invalid logging level: {new}')

        else:
            self.__announcement_level = new

    @property
    def announcement_text(self) -> str:
        """
        The announcement text.

        Returns:
            str:
                The announcement text.
        """
        if self.__announcement_text is None:
            raise ValueError('Announcement text has not been set. Try calling `replace_placeholders()` first.')

        return self.__announcement_text

    @announcement_text.setter
    def announcement_text(self, value: str) -> None:
        """
        Set the announcement text.

        The announcement text is the text that will be logged when the announcement is announced.

        Parameters:
            value (str):
                The announcement text

        Returns:
            None
        """
        self.__announcement_text = value

    @property
    def level(self):
        """
        The logging level for the announcement.

        Returns:
            str:
                The logging level for the announcement.
        """
        return self.announcement_level

    @property
    def owner(self) -> InspyLogger:
        """
        The owner of the announcement.
        """
        return self.__owner

    @owner.setter
    @validate_type(InspyLogger)
    def owner(self, value: InspyLogger) -> None:
        """
        Set the owner of the announcement.

        Parameters:
            value (InspyLogger):
                The owner of the announcement.

        Returns:
            None
        """
        self.__owner = value

    @property
    def replaced(self) -> bool:
        """
        The replacement status of the announcement. True if the placeholders have been replaced, False otherwise.

        Returns:
            bool:
                The replacement status of the announcement.
        """
        return self.__replaced

    @replaced.setter
    @validate_type(bool)
    def replaced(self, new: bool) -> None:
        """
        Set the replacement status of the announcement.

        Parameters:
            new (bool):
                The replacement status of the announcement.

        Returns:
            None

        Raises:
            ValueError:
                If the announcement has already been replaced.
        """
        self.__replaced = new

    @property
    def template(self) -> str:
        """
        The template for the announcement. A string with placeholders for the announcement.

        Note:
            The placeholders in the template are replaced with the values of the placeholders in the announcement.
            Here's an example of a template:

                ```python
                'Initialized {name}.'
                ```

            In this example, `{name}` is a placeholder that will be replaced with the value of the built-in `name`
            placeholder, which is the name of the owner of the announcement.

        Returns:
            str:
                The template for the announcement.

        """
        return self.__template

    @template.setter
    @validate_type(str)
    def template(self, new) -> None:
        """
        Set the template for the announcement.

        Args:
            new (str):
                The template for the announcement. A string with placeholders for the announcement, e.g.
                'Initialized {name}.'

        Returns:
            None

        Raises:
            ValueError:
                If the announcement has already been announced, though this exception gets caught and a warning is
                issued instead.

            TypeError:
                If the template is not a string.
        """
        if self.announced:
            try:
                raise ValueError('Announcement has already been announced. Property `template` remains unchanged!')
            except ValueError as e:
                warn(e.__str__())

        self.__template = new

    def announce(self):
        """
        Announce the announcement.
        """
        self.replace_placeholders()
        if self.announced:
            warn('Announcement has already been announced. Skipping announcement.')
            return

        if hasattr(self.owner, self.level.lower()):
            getattr(self.owner, self.level.lower())(str(self))
        else:
            self.owner.warning(f'Invalid logging level: {self.level}')
            self.owner.debug(str(self))

        self.announced = True

    def replace_placeholders(self):
        """
        Replace the placeholders in the announcement text with the values of the placeholders in the announcement. The
        placeholders are replaced in the announcement text, which is the text that will be logged when the announcement
        is announced.

        Placeholders are replaced in the announcement text by calling the `filled_string` property of the placeholders,
        which returns the value of the placeholders in the announcement.

        Here's an example of a placeholder:

            ```python
            '{name}'
            ```

        Returns:
            None
        """
        if not self.replaced:
            self.announcement_text = self.placeholders.filled_string
            self.replaced = True

    def __str__(self):

        return self.placeholders.filled_string



class ByteConverter:
    """
    A class to convert between various units of digital storage.

    Supported units: Bytes, Kilobytes, Megabytes, Gigabytes, Terabytes, Petabytes, Exabytes, Zettabytes, Yottabytes,
                     Kibibytes, Mebibytes, Gibibytes, Tebibytes, Pebibytes, Exbibytes, Zebibytes, Yobibytes.

    Methods:
        convert(value, from_unit, to_unit): Converts the given value from one unit to another.
    """

    # Dictionary to store conversion rates relative to bytes
    _conversion_rates = {
        'bytes': 1,
        'kilobytes':  1000,
        'megabytes':  1000 ** 2,
        'gigabytes':  1000 ** 3,
        'terabytes':  1000 ** 4,
        'petabytes':  1000 ** 5,
        'exabytes':   1000 ** 6,
        'zettabytes': 1000 ** 7,
        'yottabytes': 1000 ** 8,
        'kibibytes':  1024,
        'mebibytes':  1024 ** 2,
        'gibibytes':  1024 ** 3,
        'tebibytes':  1024 ** 4,
        'pebibytes':  1024 ** 5,
        'exbibytes':  1024 ** 6,
        'zebibytes':  1024 ** 7,
        'yobibytes':  1024 ** 8
    }

    @classmethod
    def convert(cls, value: float, from_unit: str, to_unit: str) -> float:
        """
        Converts the given value from one unit to another.

        Parameters:
            value (float): The value to convert.
            from_unit (str): The unit to convert from. Must be one of the supported units.
            to_unit (str): The unit to convert to. Must be one of the supported units.

        Returns:
            float: The converted value in the target unit.

        Raises:
            ValueError: If an unsupported unit is specified.
        """
        from_unit = from_unit.lower()
        to_unit = to_unit.lower()

        if from_unit not in cls._conversion_rates or to_unit not in cls._conversion_rates:
            raise ValueError("Unsupported unit. Supported units are: " + ', '.join(cls._conversion_rates.keys()))

        # Convert the value to bytes first, then to the target unit
        value_in_bytes = value * cls._conversion_rates[from_unit]
        return value_in_bytes / cls._conversion_rates[to_unit]



"""


Author:
    Inspyre Softworks

Project:
    inSPy-Logger

File:
    inspy_logger/constants.py


Description:


"""


import logging
from rich.logging import RichHandler

DEFAULT_LOGGING_LEVEL = logging.DEBUG

HANDLER_TYPES = {
        'console': RichHandler,
        'file': logging.FileHandler
        }

LEVEL_MAP = {
    'internal': 5,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
    'fatal': logging.FATAL,
}
"""A mapping of level names to their corresponding logging levels."""


LEVELS = [level.upper() for level in LEVEL_MAP]

INTERNAL = LEVEL_MAP['debug'] - 5


INTERACTIVE_SESSION = __name__ != '__main__'
"""A flag to indicate whether the session is interactive."""



from logging import Handler
import logging


class BufferingHandler(Handler):

    def __init__(self):
        super().__init__()
        self.buffer = []
        self.replaying = False

    def emit(self, record):
        if not self.replaying:
            self.buffer.append(record)

    def replay_logs(self, logger):
        """
        Replays the buffered logs and restores the original logging level of the logger.

        Parameters:
            logger:
                The logger instance to replay the logs to.

        Returns:
            None

        Examples:
            >>> handler = BufferingHandler()
            >>> logger = logging.getLogger()
            >>> logger.addHandler(handler)
            >>> logger.setLevel(logging.DEBUG)
            >>> logger.debug("Debug message")
            >>> logger.info("Info message")
            >>> logger.warning("Warning message")
            >>> handler.replay_logs(logger)
            # The buffered logs are replayed to the logger and the original logging level is restored.
        """

        self.replaying = True
        orig_level = logger.level
        logger.setLevel(logging.CRITICAL)

        for record in self.buffer:
            logger.handle(record)

        self.buffer.clear()

        self.replaying = False

        logger.setLevel(orig_level)



import inspect
from inspy_logger import LOG_DEVICE, Logger


def _determine_calling_method(instance):
    """
    Attempt to determine the name of the method calling through `instance`.
    This function inspects the call stack for a frame where `self` is `instance`.
    Returns:
        str: The method name (string) of the calling method.
    Raises:
        RuntimeError: If no calling method can be identified.
    """
    stack = inspect.stack()
    # Start from 1 to skip the current frame
    for frame_record in stack[1:]:
        if frame_record.frame.f_locals.get('self') is instance:
            return frame_record.function
    raise RuntimeError("Could not determine the calling method's name.")


def _get_parent_logging_device():
    """
    Determines the parent logging device by inspecting the caller's local variables.
    It checks for 'logger' and 'parent_log_device' in the caller's locals. If found,
    returns it as the parent logger.

    Returns:
        Logger: The parent logging device.

    Raises:
        ValueError: If neither 'logger' nor 'parent_log_device' is found in the caller's locals.
    """
    LOG_DEVICE.debug("Determining parent logging device")

    caller_frame = inspect.currentframe().f_back
    caller_locals = caller_frame.f_locals

    if "logger" in caller_locals:
        return caller_locals["logger"]
    elif "parent_log_device" in caller_locals:
        return caller_locals["parent_log_device"]
    else:
        raise ValueError(
            "Unable to determine the parent logging device. "
            "Expected either 'logger' or 'parent_log_device' in caller's locals."
        )


class LoggableDescriptor:
    """
    A descriptor that, when accessed from an instance of a class inheriting from 'Loggable',
    returns a logger specific to the calling method.

    This relies on the instance having a 'create_logger' method and a class-level 'class_logger'.
    """

    def __get__(self, instance, owner):
        if instance is None:
            # Accessing through the class, return class-level logger
            return owner.class_logger

        method_name = _determine_calling_method(instance)
        # Get a child logger named after the class and method
        return instance.create_logger(name=method_name, is_method=True)


class Loggable:
    """
    A base class that provides logging capabilities to classes that inherit from it.

    It obtains a parent logger (either from a given parent_log_device or by introspection)
    and sets up a class-level logger. Instances can create method-specific or child loggers
    easily.

    Attributes:
        parent_log_device (Logger, optional): A parent logging device used to initialize this logger.
        class_logger (Logger, class-level): A class-level logger shared across all instances.
    """
    method_logger = LoggableDescriptor()
    class_logger = None

    def __init__(self, parent_log_device=None, **kwargs):
        """
        Initialize the Loggable instance.

        Parameters:
            parent_log_device (Logger, optional): The parent log device, if available.
            **kwargs: Additional keyword arguments.

        The instance-level logger is derived either from the given parent_log_device or by
        using _get_parent_logging_device(). The class-level logger is also set if not already done.
        """
        self.parent_log_device = parent_log_device
        self.__log_name = self.__class__.__name__

        if self.parent_log_device is not None:
            self.__log_device = self.parent_log_device.get_child(self.__log_name)
        else:
            self.__log_device = _get_parent_logging_device().get_child(self.__log_name)

        # Set up class-level logger if it's not already set
        if self.__class__.class_logger is None:
            self.__class__.class_logger = self.__log_device

    @property
    def log_device(self):
        """Return the instance's log_device."""
        return self.__log_device

    @log_device.setter
    def log_device(self, new):
        """Set a new logger device."""
        if not isinstance(new, Logger):
            raise TypeError('log_device must be of type "Logger"')
        self.__log_device = new

    def create_child_logger(self, name=None, override=False, is_method=False, **kwargs):
        """
        Create and return a child logger of this instance's logger.

        Parameters:
            name (str, optional): Name of the child logger. If None, caller's name is used.
            override (bool, optional): Override membership checks if supported by the logger.
            is_method (bool, optional): Flag indicating that this logger is tied to a method.

        Returns:
            Logger: A child logger instance.
        """
        if name is None:
            # If no name provided, try to get the caller's function name
            name = inspect.stack()[1].function

        kwargs['is_method'] = is_method
        return self.class_logger.get_child(name, override=override, **kwargs)

    def create_logger(self, name=None, is_method=False, **kwargs):
        """
        Create a logger for a given name, defaults to the caller's function name if not provided.

        Parameters:
            name (str, optional): Logger name.
            is_method (bool, optional): Flag indicating that this logger is tied to a method.

        Returns:
            Logger: A new logger instance.
        """
        if name is None:
            name = inspect.stack()[1].function
        return self.create_child_logger(name=name, is_method=is_method, **kwargs)

    def __is_member__(self):
        """
        Checks whether the caller of this method is a member of the same class.
        If not, raises a PermissionError.

        This uses frame inspection to ensure that 'self' in the caller frame
        is an instance of the same class.
        """
        log_device = self.log_device.get_child("__is_member__")
        log = log_device

        current_frame = inspect.currentframe()
        caller_frame = current_frame.f_back
        caller_self = caller_frame.f_locals.get("self", None)

        log.internal("Checking if caller is a member of this class...")
        if not isinstance(caller_self, self.__class__):
            raise PermissionError(
                f"Access denied. Method can only be accessed by members of the same class. "
                f"{type(caller_self).__name__} is not a valid member."
            )

        log.internal(f"Access granted to {caller_self.__class__.__name__}")


import platform
import multiprocessing
import psutil


# Function to check if the system architecture is 64-bit
def is_64bit():
    """
    Checks if the system architecture is 64-bit.

    Returns:
        bool: True if the system architecture is 64-bit, False otherwise.
    """
    return platform.architecture()[0] == '64bit'


# Function to get the number of cores
def number_of_cores():
    """
    Gets the number of cores in the system.

    Returns:
        int: The number of cores in the system.
    """
    return multiprocessing.cpu_count()


# Function to get the processor core information
def get_processor_core_info():
    """
    Gets the processor core information.

    Returns:
        dict: A dictionary containing the total number of cores, the number of physical cores, and the number of logical cores.
    """
    phys = psutil.cpu_count(logical=False)
    total = number_of_cores()

    return {
        'total': total,
        'physical': phys,
        'logical': total - phys
    }


# Function to get the processor information
def _get_processor_info():
    """
    Gets the processor information.

    Returns:
        dict: A dictionary containing the name of the processor, the architecture of the processor, and the core information of the processor.
    """
    cores = get_processor_core_info()

    return {
        'name': platform.processor(),
        'architecture': '64-bit' if is_64bit() else '32-bit',
        'cores': get_processor_core_info()
    }


# A dictionary containing the processor information
PROCESSOR_INFO = _get_processor_info()
